@@ -1,3 +1,7 @@
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #ifndef _COMMON_H
 #define _COMMON_H
 
@@ -30,6 +34,13 @@ typedef enum {
 } dbus_session_type;
 #endif
 
+typedef enum {
+  TOE_normal,
+  TOE_emergency,
+  TOE_dbus // a request was made on a D-Bus interface (the native D-Bus or MPRIS interfaces)-- don't
+           // wait for the dbus thread to exit
+} type_of_exit_type;
+
 #define sps_extra_code_output_stalled 32768
 #define sps_extra_code_output_state_cannot_make_ready 32769
 
@@ -62,6 +73,7 @@ typedef enum {
 typedef enum {
   VCP_standard = 0,
   VCP_flat,
+  VCP_dasl_tapered,
 } volume_control_profile_type;
 
 typedef enum {
@@ -111,7 +123,14 @@ typedef struct {
   config_t *cfg;
   int endianness;
   double airplay_volume; // stored here for reloading when necessary
-  char *appName;         // normally the app is called shairport-syn, but it may be symlinked
+  double default_airplay_volume;
+  double high_threshold_airplay_volume;
+  uint64_t last_access_to_volume_info_time;
+  int limit_to_high_volume_threshold_time_in_minutes; // revert to the high threshold volume level
+                                                      // if the existing volume level exceeds this
+                                                      // and hasn't been used for this amount of
+                                                      // time (0 means never revert)
+  char *appName; // normally the app is called shairport-syn, but it may be symlinked
   char *password;
   char *service_name; // the name for the shairport service, e.g. "Shairport Sync Version %v running
                       // on host %h"
@@ -120,10 +139,18 @@ typedef struct {
   char *pa_server;           // the pulseaudio server address that Shairport Sync will play on.
   char *pa_application_name; // the name under which Shairport Sync shows up as an "Application" in
                              // the Sound Preferences in most desktop Linuxes.
-  // Defaults to "Shairport Sync". Shairport Sync must be playing to see it.
+  // Defaults to "Shairport Sync".
 
   char *pa_sink; // the name (or id) of the sink that Shairport Sync will play on.
 #endif
+#ifdef CONFIG_PW
+  char *pw_application_name;  // the name under which Shairport Sync shows up as an "Application" in
+                              // the Sound Preferences in most desktop Linuxes.
+                              // Defaults to "Shairport Sync".
+
+  char *pw_node_name; // defaults to the application's name, usually "shairport-sync".
+  char *pw_sink_target; // leave this unset if you don't want to change the sink_target.
+#endif
 #ifdef CONFIG_METADATA
   int metadata_enabled;
   char *metadata_pipename;
@@ -131,6 +158,7 @@ typedef struct {
   int metadata_sockport;
   size_t metadata_sockmsglength;
   int get_coverart;
+  double metadata_progress_interval; // 0 means no progress reports
 #endif
 #ifdef CONFIG_MQTT
   int mqtt_enabled;
@@ -147,19 +175,22 @@ typedef struct {
   int mqtt_publish_parsed;
   int mqtt_publish_cover;
   int mqtt_enable_remote;
+  char *mqtt_empty_payload_substitute;
 #endif
-  uint8_t hw_addr[6];
+  uint8_t ap1_prefix[6];
+  uint8_t hw_addr[8]; // only needs 6 but 8 is handy when converting this to a number
   int port;
   int udp_port_base;
   int udp_port_range;
   int ignore_volume_control;
   int volume_max_db_set; // set to 1 if a maximum volume db has been set
   int volume_max_db;
-  int no_sync;            // disable synchronisation, even if it's available
-  int no_mmap;            // disable use of mmap-based output, even if it's available
-  double resyncthreshold; // if it get's out of whack my more than this number of seconds, resync.
-                          // Zero means never
-                          // resync.
+  int no_sync;                 // disable synchronisation, even if it's available
+  int no_mmap;                 // disable use of mmap-based output, even if it's available
+  double resync_threshold;     // if it gets out of whack by more than this number of seconds, do a
+                               // resync. if zero, never do a resync.
+  double resync_recovery_time; // if sync is late, drop the delay but also drop the following frames
+                               // up to the resync_recovery_time
   int allow_session_interruption;
   int timeout; // while in play mode, exit if no packets of audio come in for more than this number
                // of seconds . Zero means never exit.
@@ -205,9 +236,11 @@ typedef struct {
   // char *errfile;
   char *configfile;
   char *regtype; // The regtype is the service type followed by the protocol, separated by a dot, by
-                 // default “_raop._tcp.”.
-  char *interface;     // a string containg the interface name, or NULL if nothing specified
-  int interface_index; // only valid if the interface string is non-NULL
+                 // default “_raop._tcp.” for AirPlay 1.
+  char *regtype2;  // The regtype is the service type followed by the protocol, separated by a dot,
+                   // by default “_raop._tcp.” for AirPlay 2.
+  char *interface; // a string containg the interface name, or NULL if nothing specified
+  int interface_index;                        // only valid if the interface string is non-NULL
   double audio_backend_buffer_desired_length; // this will be the length in seconds of the
                                               // audio backend buffer -- the DAC buffer for ALSA
   double audio_backend_buffer_interpolation_threshold_in_seconds; // below this, soxr interpolation
@@ -289,14 +322,21 @@ typedef struct {
                    // can't use IP numbers as they might be given to different devices
                    // can't get hold of MAC addresses.
                    // can't define the nvll linked list struct here
+
+#ifdef CONFIG_AIRPLAY_2
+  uint64_t airplay_features;
+  uint32_t airplay_statusflags;
+  char *airplay_device_id; // for the Bonjour advertisement and the GETINFO PList
+  char *airplay_pin;       // non-NULL, 4 char PIN, if required for pairing
+  char *airplay_pi;        // UUID in the Bonjour advertisement and the GETINFO Plist
+  char *nqptp_shared_memory_interface_name; // client name for nqptp service
+#endif
+  int unfixable_error_reported; // only report once.
 } shairport_cfg;
 
-// accessors to config for multi-thread access
-double get_config_airplay_volume();
-void set_config_airplay_volume(double v);
-
-uint32_t nctohl(const uint8_t *p); // read 4 characters from *p and do ntohl on them
-uint16_t nctohs(const uint8_t *p); // read 2 characters from *p and do ntohs on them
+uint32_t nctohl(const uint8_t *p);  // read 4 characters from *p and do ntohl on them
+uint16_t nctohs(const uint8_t *p);  // read 2 characters from *p and do ntohs on them
+uint64_t nctoh64(const uint8_t *p); // read 8 characters from *p to a uint64_t
 
 void memory_barrier();
 
@@ -356,16 +396,28 @@ char *base64_enc(uint8_t *input, int length);
 uint8_t *rsa_apply(uint8_t *input, int inlen, int *outlen, int mode);
 
 // given a volume (0 to -30) and high and low attenuations in dB*100 (e.g. 0 to -6000 for 0 to -60
-// dB), return an attenuation depending on a linear interpolation along along the range
+// dB), return an attenuation depending on a linear interpolation along the range
 double flat_vol2attn(double vol, long max_db, long min_db);
 
+// The intention behind dasl_tapered is that a given percentage change in volume should result in
+// the same percentage change in perceived loudness. For instance, doubling the volume level should
+// result in doubling the perceived loudness. With the range of AirPlay volume being from -30 to 0,
+// doubling the volume from -22.5 to -15 results in an increase of 10 dB. Similarly, doubling the
+// volume from -15 to 0 results in an increase of 10 dB. For compatibility with mixers having a
+// restricted attenuation range (e.g. 30 dB), "dasl_tapered" will switch to a flat profile at low
+// AirPlay volumes.
+double dasl_tapered_vol2attn(double vol, long max_db, long min_db);
 // given a volume (0 to -30) and high and low attenuations in dB*100 (e.g. 0 to -6000 for 0 to -60
 // dB), return an attenuation depending on the transfer function
 double vol2attn(double vol, long max_db, long min_db);
 
-// return a monolithic (always increasing) time in nanoseconds
-// uint64_t get_absolute_time_in_fp(void); // obselete
-uint64_t get_absolute_time_in_ns(void);
+// return a time in nanoseconds
+#ifdef COMPILE_FOR_LINUX_AND_FREEBSD_AND_CYGWIN_AND_OPENBSD
+// Not defined for macOS
+uint64_t get_realtime_in_ns(void);
+#endif
+uint64_t get_absolute_time_in_ns(void);  // monotonic_raw or monotonic
+uint64_t get_monotonic_time_in_ns(void); // NTP-disciplined
 
 // time at startup for debugging timing
 extern uint64_t ns_time_at_startup, ns_time_at_last_debug_message;
@@ -374,12 +426,9 @@ extern uint64_t ns_time_at_startup, ns_time_at_last_debug_message;
 
 uint32_t uatoi(const char *nptr);
 
-// this is for allowing us to cancel the whole program
-extern pthread_t main_thread_id;
-
 extern shairport_cfg config;
 extern config_t config_file_stuff;
-extern int emergency_exit;
+extern int type_of_exit_cleanup; // normal, emergency, dbus requested...
 
 int config_set_lookup_bool(config_t *cfg, char *where, int *dst);
 
@@ -402,8 +451,8 @@ extern pthread_mutex_t the_conn_lock;
   pthread_mutex_unlock(&the_conn_lock);
 
 // wait for the specified time in microseconds -- it checks every 20 milliseconds
-int sps_pthread_mutex_timedlock(pthread_mutex_t *mutex, useconds_t dally_time,
-                                const char *debugmessage, int debuglevel);
+// int sps_pthread_mutex_timedlock(pthread_mutex_t *mutex, useconds_t dally_time,
+//                                 const char *debugmessage, int debuglevel);
 // wait for the specified time, checking every 20 milliseconds, and block if it can't acquire the
 // lock
 int _debug_mutex_lock(pthread_mutex_t *mutex, useconds_t dally_time, const char *mutexName,
@@ -447,4 +496,44 @@ int string_update_with_size(char **str, int *flag, char *s, size_t len);
 // from https://stackoverflow.com/questions/13663617/memdup-function-in-c, with thanks
 void *memdup(const void *mem, size_t size);
 
+int bind_socket_and_port(int type, int ip_family, const char *self_ip_address, uint32_t scope_id,
+                         uint16_t *port, int *sock);
+
+uint16_t bind_UDP_port(int ip_family, const char *self_ip_address, uint32_t scope_id, int *sock);
+
+void socket_cleanup(void *arg);
+void mutex_unlock(void *arg);
+void rwlock_unlock(void *arg);
+void mutex_cleanup(void *arg);
+void cv_cleanup(void *arg);
+void thread_cleanup(void *arg);
+#ifdef CONFIG_AIRPLAY_2
+void plist_cleanup(void *arg);
+#endif
+
+char *debug_malloc_hex_cstring(void *packet, size_t nread);
+
+// from https://stackoverflow.com/questions/13663617/memdup-function-in-c, with thanks
+// allocates memory and copies the content to it
+// analogous to strndup;
+void *memdup(const void *mem, size_t size);
+
+// the difference between two unsigned 32-bit modulo values as a signed 32-bit result
+// now, if the two numbers are constrained to be within 2^(n-1)-1 of one another,
+// we can use their as a signed 2^n bit number which will be positive
+// if the first number is the same or "after" the second, and
+// negative otherwise
+
+int32_t mod32Difference(uint32_t a, uint32_t b);
+
+int get_device_id(uint8_t *id, int int_length);
+
+#ifdef CONFIG_USE_GIT_VERSION_STRING
+extern char git_version_string[];
+#endif
+
 #endif // _COMMON_H
+
+#ifdef __cplusplus
+}
+#endif
